static char *sccsid = "@(#) SupportPacks/ms0c/public/mqsecche.c, supportpacks, p000, p000-L000330          1.5  00/03/06 16:53:59";
/*---------------------------------------------------------------------
 *
 *  Author(s)     : Benjamin Y. Goldshlager
 *
 *  Name: MQSECCHE.C
 *
 *  Description:
 *      MQ Security Channel Exit main entry routines.
 *      This code is base on amqdcexc.c sample
 *
 *
 *  Notes:
 *
 *  Change Log:
 *    $Log$
 *
 *  Who                       Date     Defect Description
 *  -----------------------   -------- ------ -----------
 *  Benjamin Y. Goldshlager   11/30/98        Initial Version
 *  Benjamin Y. Goldshlager   04/13/99        Add pParms to sealData() and
 *                                              unSealData()
 *--------------------------------------------------------------------*/

/**********************************************************************/
/*                                                                    */
/* Module Name: MQSECCHE.C                                            */
/* MQ Security CHannel Exit based on IBM's AMQSDSC0.C Sample          */
/*                                                                    */
/* Description: GSS Channel Exit routines                             */
/*                                                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*  This file contains the DCE and Entrust GSS Channel exit routines  */
/*   for MQ.  The DCE code has NOT been tested.                       */
/*                                                                    */
/*  This module contains two main entry points :                      */
/*                                                                    */
/*    MQENTRY GSS_SEC_SCY_CHANNELEXIT (Security)                      */
/*    MQENTRY GSS_SEC_SRM_CHANNELEXIT (Send/Receive/Message)          */
/*                                                                    */
/*  These functions call the various internal routines depending on   */
/*  the reason they were invoked.                                     */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Notes:                                                             */
/*   Language:     C                                                  */
/*                                                                    */
/*   This source corresponds to the amqrdsc0 object code except that  */
/*   all internal MQSeries function calls have been stripped out.     */
/*                                                                    */
/*   On Windows 95 MQAT==MQAT_WINDOWS_NT.                             */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* User actions:                                                      */
/*                                                                    */
/*   MUST DO (PC platforms):                                          */
/*   -----------------------                                          */
/*                                                                    */
/*   Entrust:                                                         */
/*   User need to configure an INI file.  See example in MQSECCHE.INI.*/
/*                                                                    */
/*                                                                    */
/**********************************************************************/
#include "mqsecche.h"

/**********************************************************************/
/*                                                                    */
/* Function Name: GSS_SEC_SCY_CHANNELEXIT                             */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Main entry point for channel security exit               */
/*           Check to se that this has been called as a security exit */
/*           Then switch to whatever is defined by pParms->ExitReason */
/*                                                                    */
/* Input Parameters: pAgentBufferLength  UNUSED                       */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters: pChannelExitParms   Channel exit parameter block */
/*                   pChannelDefinition  Channel definition           */
/*                   pDataLength         Length of data               */
/*                   pAgentBuffer        Agent buffer                 */
/*                   pExitBufferLength   Length of exit buffer        */
/*                   pExitBufferAddr     Address of exit buffer       */
/*                                                                    */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/
#if MQAT_DEFAULT == MQAT_WINDOWS_NT
__declspec(dllexport)  void MQENTRY GSS_SEC_SCY_CHANNELEXIT (
#else
void MQENTRY GSS_SEC_SCY_CHANNELEXIT (
#endif
    PMQVOID  pChannelExitParms,   /* Channel exit parameter block */
    PMQVOID  pChannelDefinition,  /* Channel definition           */
    PMQLONG  pDataLength,         /* Length of data               */
    PMQLONG  pAgentBufferLength,  /* Length of agent buffer       */
    PMQVOID  pAgentBuffer,        /* Agent buffer                 */
    PMQLONG  pExitBufferLength,   /* Length of exit buffer        */
    PMQPTR   pExitBufferAddr)     /* Address of exit buffer       */
{
  PMQCXP                  pParms = (PMQCXP)pChannelExitParms;
  PMQCD                   pChDef = (PMQCD)pChannelDefinition;

  PSECEXITDATA    pSecData;
  PCONTEXTDATA    pContext;
  PEXITUSERAREA   pExitUserArea;
  PEXITLOCALVARS  pExitLocalVars;

  char                    szLocalQMgrName[MQ_Q_MGR_NAME_LENGTH+1];
  char                    szChannelName[MQ_MCA_NAME_LENGTH+1];

  pContext = (PCONTEXTDATA)&(pChDef->SecurityUserData);
  pExitUserArea = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars = pExitUserArea->pExitLocalVars; // Very first time this should be NULL

  dtkDebug (pExitLocalVars, "GSS_SEC_SCY_CHANNELEXIT(): Received control "
    "with pParms->ExitId \"%s (%ld)\", pParms->ExitReason \"%s (%ld)\"\n",
    getExitIdStr (pParms->ExitId), pParms->ExitId,
    getExitReasonStr (pParms->ExitReason), pParms->ExitReason);


  /* Invoked as security exit? */
  if (!(SECURITYEXITCALL))
  {
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> GSS_SEC_SCY_CHANNELEXIT(): "
	  "Unexpected call request in pParms->ExitId=<%ld>\n",
	  __FILE__, __LINE__, pParms->ExitId, getExitIdStr (pParms->ExitId));

    pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
    return;
  }
  else
  {
    /* free buffer used to pass back message on previous exit call */
    if ((pExitBufferAddr != NULL) && (*pExitBufferAddr != NULL))
    {
      dtkDebug (pExitLocalVars, "GSS_SEC_SCY_CHANNELEXIT(): Freeing pExitBufferAddr "
	"of size <%d> bytes\n", *pExitBufferLength);
      free (*pExitBufferAddr);
      *pExitBufferAddr = NULL;
      *pExitBufferLength = 0;
    } /* endif */
  } /* endif */

  /* If OS/2: turn this thread into a pthread */
#if MQAT==MQAT_OS2
   pthread_pseudo_thread_base_start();
#endif

     /*******************************************************************/
     /* now switch to what ever function we were called to do           */
     /*******************************************************************/
  switch (pParms->ExitReason)
  {
    case MQXR_INIT:
      convertMQCHARToString (szChannelName,
	           pChDef->ChannelName,
	           MQ_MCA_NAME_LENGTH);

      memset (pContext, 0, sizeof (CONTEXTDATA));

      pParms->ExitResponse =
	initSecurityExit (&(pParms->ExitData),
	          szChannelName,
	          (PEXITUSERAREA)&(pParms->ExitUserArea),
	          (PCONTEXTDATA)&(pChDef->SecurityUserData));
      // initSecurityExit() allocated heap memory for SECEXITDATA
      // and EXITLOCALVARS
      if (pParms->ExitResponse == MQXCC_OK)
      {
	pSecData = pExitUserArea->pSecExitData;
	pExitLocalVars = pExitUserArea->pExitLocalVars;

	convertMQCHARToString (pExitLocalVars->szLocalQMgrName,
	             pChDef->QMgrName,
	             MQ_Q_MGR_NAME_LENGTH);
	convertMQCHARToString (pExitLocalVars->szChannelExitINIFile,
	             pParms->ExitData,
	             MQ_EXIT_DATA_LENGTH);

	dtkDisplay (pExitLocalVars, "GSS_SEC_SCY_CHANNELEXIT(): <======  <%s> is "
	      "initiating channel <%s>  ======>\n",
	      pExitLocalVars->szLocalQMgrName, szChannelName);

#ifdef GSS_ENTRUST_API
	pParms->ExitResponse = initEntrustEnv (pExitLocalVars);
	if (pParms->ExitResponse == MQXCC_OK)
#endif
	  /* Could be MQXCC_OK||MQXCC_SUPPRESS_FUNCTION||MQXCC_SEND_AND_REQUEST_SEC_MSG etc.*/
	  EXPECTSECMSG(pParms->ExitResponse,
	       pSecData,
	       GSS_SEC_INIT_SEC); /* exprected message in both Sender/Receiver */
      }

      break;

    case MQXR_INIT_SEC:     /* Receiver initializes */
      pSecData = pExitUserArea->pSecExitData;
      pExitLocalVars = pExitUserArea->pExitLocalVars;

      /* if necessary and possible check target name against partner name in exit parmeters */
      pParms->ExitResponse =
	compareMQNames (pExitLocalVars,
	        pExitLocalVars->mqTargetName,
	&(pParms->PartnerName[0]),
	pChDef->ChannelType,
	(PMQCHAR)&(pChDef->ChannelName),
	        (PMQCHAR)&(pChDef->QMgrName));

      if (pParms->ExitResponse == MQXCC_OK)
      {
	pParms->ExitResponse =
	  savePartnerName (pExitUserArea,
	          &(pSecData->PartnerGSSName),
	            (PMQCHAR48)&(pParms->PartnerName),
	            (PMQBYTE)&(pChDef->ChannelName));

	if (pParms->ExitResponse == MQXCC_OK)
	{
	  pParms->ExitResponse =
	    createCreds (pParms,
	           pChDef,
	           &(pSecData->credentialHandle),
	           &(pContext->contextHandle));

	  if (pParms->ExitResponse == MQXCC_OK)
	  {
	    pParms->ExitResponse =
	      setKeyEstblishmentProtocol (pExitLocalVars,
	                    pSecData->credentialHandle);

	    if (pParms->ExitResponse == MQXCC_OK)
	    {
	      initSecContext (pParms,
	              pSecData,
	              pContext,
	              NULL,
	              0,
	              (PMQBYTE *)pExitBufferAddr,
	              pExitBufferLength,
	              (PMQCHAR)&(pChDef->ChannelName));

	      EXPECTSECMSG(pParms->ExitResponse,
	             pSecData,
	             GSS_SEC_ACC_SEC);
	    }
	  } /* endif */
	} /* endif */
      } /* endif */

      break;

    case MQXR_SEC_MSG:      /*
	       *  Sender accepts (GSS_SEC_ACC_SEC).
	       *  Receiver complete initizliation (GSS_SEC_INIT_SEC)
	       */
      processSecurityMessage (pParms,
	          pChDef,
	    pAgentBuffer,
	    *pDataLength,
	    pExitBufferAddr,
	    pExitBufferLength);
      break;

    case MQXR_TERM:
      pSecData = pExitUserArea->pSecExitData;

      convertMQCHARToString (szChannelName,
	           pChDef->ChannelName,
	           MQ_MCA_NAME_LENGTH);

      /* In case initSecExit failed to initalize pExitLocalVars->szLocalQMgrName */
      convertMQCHARToString (szLocalQMgrName,
	           pChDef->QMgrName,
	           MQ_Q_MGR_NAME_LENGTH);

      dtkDisplay (pExitLocalVars, "GSS_SEC_SCY_CHANNELEXIT(): <======  <%s> is "
	"terminating channel <%s>  ======>\n", szLocalQMgrName, szChannelName);

      terminateSecurityExit (pParms,
	   pSecData,
	           pContext);

      /* terminateSecurityExit freed our local variables - must reflect this */
      pExitLocalVars = NULL;

#ifndef ENTRUST_BUG_FIXED
#if MQAT_DEFAULT==MQAT_UNIX
      sleep (2);
      /* Allows Entrust DLL to cleanup before it is unloaded.. */
#else
      Sleep (2000L);
#endif
#endif
      break;

    default:
      /* Unrecognized exit function */
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> GSS_SEC_SCY_CHANNELEXIT(): "
	"Unexpected call request in pParms->ExitReason=<%ld, \"%s\">\n",
	__FILE__, __LINE__, pParms->ExitReason, getExitReasonStr (pParms->ExitReason));

      pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;

  } /* endswitch */

  /* If OS/2: turn this thread back into a system thread */
#if MQAT==MQAT_OS2
  pthread_pseudo_thread_base_end();
#endif

  if (pParms->ExitResponse == MQXCC_SUPPRESS_FUNCTION ||
    pParms->ExitResponse == MQXCC_CLOSE_CHANNEL)
  {
    pParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
  }
  else
  {
    if (pParms->ExitResponse2 == MQXR2_USE_EXIT_BUFFER)
    {
      *pDataLength = *pExitBufferLength;
    } /* endif */
  } /* end if */

  /*
   *  Only console output (if option was specified in INI file) will be performed
   *  during MQXR_TERM.
   */
  dtkDebug (pExitLocalVars, "GSS_SEC_SCY_CHANNELEXIT(): pParms->ExitReason \"%s (%ld)\" "
      "is returning pParms->ExitResponse \"%s (%ld)\"\n",
      getExitReasonStr (pParms->ExitReason), pParms->ExitReason,
      getExitResponseStr (pParms->ExitResponse), pParms->ExitResponse);
}


/**********************************************************************/
/*                                                                    */
/* Function Name: GSS_SEC_SRM_CHANNELEXIT                             */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Main entry point for channel message encryption and      */
/*           decryption: acts as a send exit, a receive exit or       */
/*           a message exit.                                          */
/*                                                                    */
/* Input Parameters: pAgentBufferLength  UNUSED                       */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters: pChannelExitParms   Channel exit parameter block */
/*                   pChannelDefinition  Channel definition           */
/*                   pDataLength         Length of data               */
/*                   pAgentBuffer        Agent buffer                 */
/*                   pExitBufferLength   Length of exit buffer        */
/*                   pExitBufferAddr     Address of exit buffer       */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/
#if MQAT_DEFAULT == MQAT_WINDOWS_NT
__declspec(dllexport) void MQENTRY GSS_SEC_SRM_CHANNELEXIT (
#else
void MQENTRY GSS_SEC_SRM_CHANNELEXIT (
#endif
    PMQVOID  pChannelExitParms,   /* Channel exit parameter block */
    PMQVOID  pChannelDefinition,  /* Channel definition           */
    PMQLONG  pDataLength,         /* Length of data               */
    PMQLONG  pAgentBufferLength,  /* Length of agent buffer       */
    PMQVOID  pAgentBuffer,        /* Agent buffer                 */
    PMQLONG  pExitBufferLength,   /* Length of exit buffer        */
    PMQPTR   pExitBufferAddr)     /* Address of exit buffer       */
{
  PMQCXP        pParms;
  PMQCD         pChDef;
  PCONTEXTDATA  pContext;

  OM_uint32              major_status = 0;
  OM_uint32              minor_status = 0;
  gss_buffer_desc        input_buffer;
  gss_buffer_desc        output_buffer = { 0, NULL };
  PMQBYTE                msg = (PMQBYTE)pAgentBuffer;

  PEXITUSERAREA   pExitUserArea;
  PEXITLOCALVARS  pExitLocalVars;

  char    szChannelName[MQ_MCA_NAME_LENGTH+1];

  pParms   = (PMQCXP)pChannelExitParms;
  pChDef   = (PMQCD)pChannelDefinition;
  pContext = (PCONTEXTDATA)&(pChDef->SecurityUserData);

  pExitUserArea = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars = pExitUserArea->pExitLocalVars;

  dtkDebug (pExitLocalVars, "GSS_SEC_SRM_CHANNELEXIT(): Received control "
    "with pParms->ExitId \"%s (%ld)\", pParms->ExitReason \"%s (%ld)\"\n",
    getExitIdStr (pParms->ExitId), pParms->ExitId,
    getExitReasonStr (pParms->ExitReason), pParms->ExitReason);

  if (!(SENDEXITCALL) && !(RECEIVEEXITCALL) && !(MSGEXITCALL))
  {
    /* Unexpectd exit request */
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> GSS_SEC_SRM_CHANNELEXIT(): "
      "Unexpected call request in pParms->ExitId=<%ld> %s\n",
      __FILE__, __LINE__, pParms->ExitId, getExitIdStr (pParms->ExitId));
    pParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
    pParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
    return;
  }
  else
  {
    if ((pExitBufferAddr != NULL) && (*pExitBufferAddr != NULL) &&
      /* On the very first call pContext is not initizlied,           */
      /* therefore, the pExitBufferAddr was not allocated by us.      */
#if GSS_DCE_API
      (pContext->mechanism == GSS_DCE_MACHANISM))
#elif GSS_ENTRUST_API
      (pContext->mechanism == GSS_ENTRUST_MECHANISM))
#else
#error No GSS_ENTRUST_API or GSS_DCE_API preprocessor defined
#endif
    {
      /* free buffer used to pass back message on previous exit call */
      dtkDebug (pExitLocalVars, "GSS_SEC_SRM_CHANNELEXIT(): Freeing pExitBufferAddr "
	"of size <%d> bytes\n", *pExitBufferLength);
      free (*pExitBufferAddr);
      *pExitBufferAddr = NULL;
      *pExitBufferLength = 0;
    } /* endif */

    if (pParms->ExitReason == MQXR_INIT)
    {
      convertMQCHARToString (szChannelName,
	           pChDef->ChannelName,
	           MQ_MCA_NAME_LENGTH);
      pParms->ExitResponse =
	initMessageExit (&(pParms->ExitData),
	         szChannelName,
	         (PEXITUSERAREA)&(pParms->ExitUserArea));
      return;
    }

    if (pParms->ExitReason == MQXR_TERM)
    {
      freeExitUserArea (pParms);
      pParms->ExitResponse = MQXCC_OK;
      return;
    } /* endif */

  } /* endif */

  /* Now test for a receive exit call with byte 9 non-zero.               */
  /* The data has not been encrypted by a                                 */
  /* matching send exit; this can                                         */
  /* happen quite validly during security context set up.                 */
  /* Such data is returned to the exit caller without further processing  */

  if (RECEIVEEXITCALL && (msg[8] != '\0')) /* zero if from send exit */
  {
    /* ExitResponse2 is passed in as MQXR2_USE_AGENT_BUFFER             */
    /* and the DataLength is unchanged.                                 */
    pParms->ExitResponse = MQXCC_OK;
    return;
  }

#if GSS_DCE_API
  if (pContext->mechanism != GSS_DCE_MACHANISM)
#elif GSS_ENTRUST_API
  if (pContext->mechanism != GSS_ENTRUST_MECHANISM)
#else
#error No GSS_ENTRUST_API or GSS_DCE_API preprocessor defined
#endif
  {       /* Correct security exit has not been run */
    /* A send exit can validly be called before the security context is   */
    /* established. Clearly we cannot encrypt without a security context. */
    /* Just return the data to the exit caller unchanged.                 */

    if (SENDEXITCALL)
    {
      *pExitBufferLength = *pDataLength;
      *pExitBufferAddr = pAgentBuffer;
      pParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;

      pParms->ExitResponse = MQXCC_OK;
      return;
    }

    /* Diagnostic output here */
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> "
      "GSS_SEC_SRM_CHANNELEXIT(): Invalid message received - "
      "pContext->mechanism <%02.2x>\n",
      __FILE__, __LINE__, pContext->mechanism);

    pParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
    pParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
    return;
  } /* endif */

  /* If OS/2: turn this thread into a pthread */
  #if MQAT==MQAT_OS2
  pthread_pseudo_thread_base_start ();
  #endif

  if (SENDEXITCALL ||
    (MSGEXITCALL && ((pChDef->ChannelType == MQCHT_SENDER) ||
    (pChDef->ChannelType == MQCHT_SERVER))))
  {       /* we are sending */
    input_buffer.length =
       (SENDEXITCALL) ? *pDataLength-8 : *pDataLength-sizeof(MQXQH);
    input_buffer.value  = &msg[(SENDEXITCALL) ? 8 : sizeof(MQXQH)];

    pParms->ExitResponse =
	sealData (pExitLocalVars,
	      pParms,
	      pContext->contextHandle,
	      &input_buffer,
	      &output_buffer);

    //if (major_status == GSS_S_COMPLETE)
    if (pParms->ExitResponse == MQXCC_OK)
    {
      /* send exit header size is 10: 8 bytes of actual header, the          */
      /* 9th byte set to zero to indicate that                               */
      /* the data has been encrypted by the send exit                        */
      /* (9th byte is never zero unless set to zero in an exit),             */
      /* 10th byte is unused to provide the correct byte alignment for htons */
      composeSecMsg (pParms,
	       (PMQBYTE *)pExitBufferAddr,
	       pExitBufferLength,
	       pAgentBuffer,
	       (SENDEXITCALL) ? 10 : sizeof(MQXQH),
	       GSS_SEC_SEAL,
	       (PMQBYTE)output_buffer.value,
	       output_buffer.length,
	       MQXCC_OK,
	       (PMQCHAR)&(pChDef->ChannelName));

      gss_release_buffer (&minor_status, &output_buffer);
    }
    else
    {       /* GSS-API call failed */
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> "
	  "GSS_SEC_SRM_CHANNELEXIT(): Could not %s message\n",
	  __FILE__, __LINE__,
	  pExitLocalVars->bEncrypt ? "Sign and Encrypt" : "Sign");
    }
  }
  else   /* we are receiving */
  {
    /* offset is 10 for receive exit: header plus 9th byte set to zero by send   */
    /* exit plus 10th byte reserved to provide correct                           */
    /* byte alignment for htons on send                                          */
    MQLONG     offset    = (RECEIVEEXITCALL) ? 10 : sizeof (MQXQH);
    PGSSSECMSG pGSS_secMsg = (PGSSSECMSG)&msg[offset];

    if (*pDataLength > (offset + (long)sizeof (pGSS_secMsg->header)))
    {
      if ((SEC_MECHANISM(pGSS_secMsg->header.mechanism)) &&
	(pGSS_secMsg->header.msgType == GSS_SEC_SEAL))
      {
	input_buffer.length = *pDataLength-offset-sizeof(pGSS_secMsg->header);
	input_buffer.value  = &(pGSS_secMsg->msg);

	pParms->ExitResponse =
	    unSealData (pExitLocalVars,
	          pParms,
	          pContext->contextHandle,
	          &input_buffer,
	          &output_buffer);

	if (pParms->ExitResponse == MQXCC_OK)
	{
	  PMQBYTE pToken;
	  /* now remove bytes 9 and 10 from header     */
	  MQLONG  headlen = (RECEIVEEXITCALL) ? 8 : sizeof(MQXQH);

	  pToken = (PMQBYTE) malloc ((size_t)(output_buffer.length+headlen));

	  if (pToken != NULL)
	  {
	    if (offset > 0)
	    {
	      memcpy (pToken, msg, (size_t)headlen);
	    }
	    memcpy (&pToken[headlen], output_buffer.value, output_buffer.length);

	    *pExitBufferAddr = pToken;
	    *pExitBufferLength = output_buffer.length + headlen;
	    pParms->ExitResponse = MQXCC_OK;
	    pParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;

	    dtkDebug (pExitLocalVars, "GSS_SEC_SRM_CHANNELEXIT(): Allocated pExitBufferAddr for "
	      "<%ld> bytes\n", *pExitBufferLength);
	    gss_release_buffer ( &minor_status, &output_buffer );
	  }
	  else    /* malloc didn't work */
	  {
	    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> "
	      "GSS_SEC_SRM_CHANNELEXIT(): Could not allocate memory\n",
	      __FILE__, __LINE__);

	    pParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
	  } /* endif malloc worked */
	}
	else
	{       /* GSS-API call failed */
	  dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> "
	    "GSS_SEC_SRM_CHANNELEXIT(): -- SECURITY VIOLATION -- "
	    "Could not Verify Signature and/or Decrypt message\n",
	    __FILE__, __LINE__);

	  //pParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
	}   /* endif ( major_status == GSS_S_COMPLETE ) */
      }
      else
      {
	/* Invalid message received */
	dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> "
	  "GSS_SEC_SRM_CHANNELEXIT(): Invalid message received - "
	  "pInMsgAddr->header.mechanism <%02.2x> pInMsgAddr->header.msgType <%ld>\n",
	  __FILE__, __LINE__, pGSS_secMsg->header.mechanism,
	  pGSS_secMsg->header.msgType);

	pParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
      } /* endif */
    }
    else
    {
      /* Invalid message length */
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> "
	  "GSS_SEC_SRM_CHANNELEXIT(): Invalid message length - Received <%d> "
	  " bytes, expecting length to be greator than <%d> bytes (offset of "
	  "10 + sizeof (GSSSECMSG))\n",
	  __FILE__, __LINE__,
	  *pDataLength, (offset + (long)sizeof (pGSS_secMsg->header)));

      pParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
    } /* endif */
  }

  /* If OS/2: turn this thread back into a system thread */
  #if MQAT==MQAT_OS2
  pthread_pseudo_thread_base_end();
  #endif

  if (pParms->ExitResponse == MQXCC_CLOSE_CHANNEL)
  //if (pParms->ExitResponse == MQXCC_CLOSE_CHANNEL ||
  //      pParms->ExitResponse == MQXCC_SUPPRESS_FUNCTION)
  {
    pParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
  }
  else
  {
    if (pParms->ExitResponse2 == MQXR2_USE_EXIT_BUFFER)
    {
      *pDataLength = *pExitBufferLength;
    } /* endif */
  } /* end if */

}


/******************************************************************/
/* Known entry point to allow it to be loadable on UNIX           */
/* This is not called by the MCAs, so it's usually a void         */
/* parameter block. But on AIX, where we're going to try a little */
/* trick to get the C++ environment setup first, having a         */
/* single entry point makes the wrapper implementation a bit      */
/* easier.                                                        */
/******************************************************************/
void MQENTRY MQStart(
    PMQVOID  pChannelExitParms,   /* Channel exit parameter block */
    PMQVOID  pChannelDefinition,  /* Channel definition           */
    PMQLONG  pDataLength,         /* Length of data               */
    PMQLONG  pAgentBufferLength,  /* Length of agent buffer       */
    PMQVOID  pAgentBuffer,        /* Agent buffer                 */
    PMQLONG  pExitBufferLength,   /* Length of exit buffer        */
    PMQPTR   pExitBufferAddr)     /* Address of exit buffer       */
{
  PMQCXP        pParms;
  PMQCD         pChDef;
  PCONTEXTDATA  pContext;
  PEXITUSERAREA   pExitUserArea;
  PEXITLOCALVARS  pExitLocalVars;

  pParms   = (PMQCXP)pChannelExitParms;
  pChDef   = (PMQCD)pChannelDefinition;
  pContext = (PCONTEXTDATA)&(pChDef->SecurityUserData);

  pExitUserArea = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars = pExitUserArea->pExitLocalVars;

  dtkDebug (pExitLocalVars, "MQStart(): Received control "
    "with pParms->ExitId \"%s (%ld)\", pParms->ExitReason \"%s (%ld)\"\n",
    getExitIdStr (pParms->ExitId), pParms->ExitId,
    getExitReasonStr (pParms->ExitReason), pParms->ExitReason);

  if (pParms->ExitId == MQXT_CHANNEL_SEC_EXIT)
  {
    GSS_SEC_SCY_CHANNELEXIT (
      pChannelExitParms,
      pChannelDefinition,
      pDataLength,
      pAgentBufferLength,
      pAgentBuffer,
      pExitBufferLength,
      pExitBufferAddr);
  }
  else
  {
    GSS_SEC_SRM_CHANNELEXIT (
      pChannelExitParms,
      pChannelDefinition,
      pDataLength,
      pAgentBufferLength,
      pAgentBuffer,
      pExitBufferLength,
      pExitBufferAddr);
  }
  return;
}
